% Matlab script like validatedPropValue but handles aliases
% Part of UMD Matlab-Slurm integration scripts
% 
% Author: Tom Payerle (payerle@umd.edu) 
% (c) University of Maryland 2022-2023
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This file is part of the git repo 
% git@gitlab.umd.edu:acigs-enduser-sw/umd-slurm-matlab-scripts.git
% Please make any and all modifications there
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function val = validatedPropValueWithAliases(ap, prop, type, defaultValue, aliases, deprecated, debug)
% Wrapper around validatedPropValue to handle aliases.  Extra args are:
%	aliases: list of alias names to be used if the primary property
%		name is not found
%	deprecated: similar to aliases, but these alias names are
%		deprecated.  I.e., a warning will be produced if they
%		are used, even if debug is off
%	debug: logical.  If set, print debug messages.
%
% NOTE: prop, and elements of aliases and deprecated MUST be double quoted
% strings.
%
% It generally behaves like validatedPropValue, but handles aliases and
% deprecated aliases for the property name.  We look at the main property
% name as well as all defined aliases/deprecated aliases, and return the
% validated Property value (via validatedPropValue) for the first named
% Property found (searching the main property name first, then aliases in
% order given, then deprecated aliases in order given).  If no named 
% Property found, returns the defaultValue (or [] if no defaultValue is given).
%
% If multiple matches found, only the first match is returned, but warnings
% will be generated for the duplicates.  If a deprecated name is the first
% match found, a deprecation warning will be generated.  If the debug flag
% is set, an debug message will be produced if the first match is not the
% main property name.
%
% Copyright 2022-2023 The University of Maryland

    narginchk(3, 7);

    if nargin < 7
        % debug defaults to false
        debug = logical(0);
    end

    if nargin < 6
        % deprecated defaults to empty list
        deprecated = [];
    end

    if nargin < 5
        % aliases defaults to empty list
        aliases = [];
    end

    if nargin < 4
        % If no defaultValue specified, use empty
        defaultValue = [];
    end

    % set allPropNames the main name plus aliases plus deprecated aliases
    allPropNames = [ [prop], aliases, deprecated ];
    numNames = length(allPropNames);
    numUndeprecated = length(aliases) + 1;
    found = logical(0);
    foundIndex = 0;

    % Loop over all PropNames
    for index = 1:numNames
        pname = allPropNames(index);
        if isprop(ap, pname)
            % pname matches a property in ap
            % is it the first match?
           if foundIndex == 0
               % Yes, first match
               foundIndex = index;
               if index > numUndeprecated
                   % Warn that matched a deprecated form
                   warning("parallelexamples:GenericSLURM:Deprecated", ...
                       "Used deprecated AdditionalProperty field %s;  " + ...
                       "please use %s instead", pname, prop);
               elseif debug && ( index > 1 )
                   % Debug mode: warn if using an alias
                   fprintf(1, "[DEBUG] Setting %s from alias %s\n", ...
		       prop, pname);
               end
           else
               % No, matched previously; warn
               tmppname = allPropNames(foundIndex);
               warning("parallelexamples:GenericSLURM:Duplicated", ...
                   "Ignoring value of Property %s, alias for %s " +  ...
                   "as already set from %s", pname, prop, tmppname);
           end
        end
    end

    % At this point, either 
    %    foundIndex == 0, meaning no matching property was found, or
    %    foundIndex > 0, meaning a matching property was found (and any 
    %       warnings re deprecation/duplication were already issued)
    if foundIndex == 0
        val = defaultValue;
    else
        pname = allPropNames(foundIndex);
        val = validatedPropValue(ap, pname, type, defaultValue);
    end
end
