% Matlab script to create a Generic Cluster profile using Slurm
% Part of UMD Matlab-Slurm integration scripts
% 
% This is an attempt to make a generic version which can be called
% from both shared and remote versions of configCluster to do most of
% the work.
%
% Author: Tom Payerle (payerle@umd.edu) 
% (c) University of Maryland 2022-2023
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% This file is part of the git repo 
% git@gitlab.umd.edu:acigs-enduser-sw/umd-slurm-matlab-scripts.git
% Please make any and all modifications there
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function cluster = configClusterCommon(Args)
    arguments
        % useLocalCluster: if set, we assume this script is running on 
	%	the cluster we are submitting jobs to
	Args.useLocalCluster (1,1) logical = logical(1)
	% interactive: if set, we assume this script is running interactively,
	%	otherwise assume we cannot ask user for input
        Args.interactive (1,1) logical = logical(1) 
	% clusterName: name of cluster to submit jobs to
        Args.clusterName = []
	% allowUnknownMatlab: if set, we will try to create profile for
	%	a Matlab version not in our list of known MATLAB versions
	%	for this cluster.  
	Args.allowUnknownMatlab  = []
	% username on cluster.  Defaults to username on current system
	Args.clusterUsername = []
	% location of plugin scripts.  If useLocalCluster is set, will 
	% default to /software/matlab/parallel_slurm/current.  Otherwise,
	% defaults to  Documents/MATLKAB/SupportPackages/$MATLAB_VERSION 
	% under home directory.  
	Args.pluginScriptDir = []
	% remoteJobStorageLocation.
	% Defaults from cluster infomation
	Args.remoteJobStorageLocation = []
	% name to use for the profile
	% Defaults to 'Clustername HPC'
	Args.profileName = []
	% Require confirmation.  Ignored unless interactive is also set
	Args.requireConfirmation (1,1) logical = logical(1)

    end

    errorIDroot = "UMD:GenericSlurm:configClusterCommon:";
    script_version = '0.1';
    interactive = Args.interactive;
    allowUnknownMatlab = Args.allowUnknownMatlab;
    useLocalCluster = Args.useLocalCluster;

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %                Get defaults
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    %--------- The cluster we are connecting to
    defaultCluster = UmdDefaultClusterName;
    clusterName = configClusterArgumentQuery( ...
       passed=Args.clusterName, ...
       prompt="Which cluster do you wish to connect to", ...
       defaultValue=defaultCluster, ...
       type="string", ...	
       interactive=interactive ...
    );
    if isempty(clusterName)
       error( errorIDroot + "NoCluster", ...
          "[ERROR] clusterName not provided and unable to default");
    end

    % Get cluster info
    cluster_info = UmdHpcClusterInfo(clusterName);
    if isempty(cluster_info)
        error( errorIDroot + "UnknownCluster", ...
	    "Unrecognized HPC cluster %s, aborting", clusterName);
    end
    ClusterName = cluster_info.ClusterName;
    cluster_login_node = cluster_info.loginNode;
    cluster_matlab_versions = cluster_info.matlabVersions;
    cluster_jobstorage_string = cluster_info.remoteJobStorageLocation;

    %--------- The version of Matlab we are running
    myMatlabRelease = version('-release'); %This returns something like 2022b
    rver = sprintf('R%s', myMatlabRelease);
    cluster_matlab_version = [];
    cluster_matlab_path = [];
    known_versions = [];
    for k = 1:numel(cluster_matlab_versions)
	vrec = cluster_matlab_versions(k);
	ver = vrec.version;
	known_versions = [ known_versions, ver];
	if ver == myMatlabRelease
		cluster_matlab_version = ver;
		cluster_matlab_path = vrec.path;
		break;
	end
    end
    if isempty(cluster_matlab_version)
	% Our matlab version does not match any known versions for 
	% this cluster.  

	% Warn user
        tmp0 = strjoin(known_versions, ", ");
        tmp1 = [ "You are running MATLAB version %s.",
		"That version is not known to this script (script version %s) ",
		"to be supported by cluster %s.  (This script believes the",
		"only versions supported by the cluster are [%s]).",
		"",
		"Please contact hpcc-help@umd.edu and provide the MATLAB ",
		"and script versions reported above.  ",
		];
        tmp2 = strjoin(tmp1);
        tmp3 = sprintf(tmp2, myMatlabRelease, script_version, ...
		clusterName, tmp0);
	warning( errorIDroot + "UnknownMATLAB", tmp3);

	% Should we continue anyway?
	tmp1 = [
	    "If you think that this MATLAB version _is_ supported by the ",
	    "cluster and this script is just out of date, please still",
	    "contact us, but you can answer yes to the question below ",
	    "and the script will try to continue.  It can create a ",
	    "profile for that version of MATLAB, but the profile will ",
	    "not work unless the correct version of MATLAB is actually ",
	    "installed on the cluster\n",
	    "Continue anyway [y/N]? ",
	];
        tmp2 = strjoin(tmp1, "\n");
	continueAnyway = configClusterArgumentQuery( ...
	    passed = allowUnknownMatlab, ...
	    interactive = interactive, ...
	    prompt = tmp2, ...
	    type = "char" ...
	    );
	% Default is no, and convert to logical
        if isempty(continueAnyway)
	    continueAnyway = logical(0);
	elseif startsWith(lower(continueAnyway), "n")
	    continueAnyway = logical(0);
	elseif startsWith(lower(continueAnyway), "y")
	    continueAnyway = logical(1);
	else
	    error( errorIDroot + "InvalidValue", ...
	        "Invalid value '%s' for continueAnyway, expecting n or y");
	end

	if continueAnyway
	    warning( errorIDroot + "UsingUnknownMATLAB", ...
	        "Continuing with cluster configuration using unrecognized MATLAB version.");
	else
	    error(errorIDroot + "UnknownMATLAB", ...
	    	"Aborting configuration due to unrecognized MATLAB version");
	end
    end %if isempty(cluster_matlab_version)

    % --------- Username (on cluster)
    % Try to get username on this system
    defUsername = getenv("USER");
    username = configClusterArgumentQuery( ...
        passed = Args.clusterUsername, ...
	defaultValue = defUsername, ...
	interactive = interactive, ...
	prompt = "Username on cluster" ...
    );


    %-------- Plugin scripts directory
    defPluginPath = [];
    if useLocalCluster
        defPluginRoot = "/software/matlab/matlab-slurm-integration/parallel_slurm/current";
	scriptMode = "shared";
    else
	% userpath already includes Documents/MATLAB components
        userdir = userpath
        rver = sprintf("R%s", myMatlabRelease);
	defPluginRoot = fullfile(userdir, "SupportPackages", rver);
	scriptMode = "nonshared";
    end

    pluginRoot = configClusterArgumentQuery( ...
        passed = Args.pluginScriptDir, ...
	defaultValue = defPluginRoot, ...
	interactive = interactive, ...
	prompt = "Please provide the path to the directory containing\n" + ...
	    "the MATLAB-Slurm integration scripts; i.e. the location\n" + ...
	    "where you untarred the MATLAB Slurm plugins into (\n" + ...
	    "it should contain a parallel subdirectory with a slurm\n" + ...
	    "subdirectory): ", ...
	type = "string" ...
    );

    if isempty(pluginRoot)
        error( errorIDroot + "NoPluginDir", ...
	   "Missing plugin directory and unable to default");
    end

    if exist(pluginRoot) ~= 7
	error( errorIDroot + "BadPluginRoot", ...
	    "MATLAB Slurm Plugin root  %s does not exist or " + ...
	    "is not a directory: %d\n", pluginRoot, exist(pluginRoot));
    end

    pluginPath = fullfile(pluginRoot, "parallel", "slurm", scriptMode);
    if exist(pluginPath) ~= 7
	error( errorIDroot + "BadPluginPath", ...
	    "MATLAB Slurm Plugin path %s does not exist or " + ...
	    "is not a directory: %d\n", pluginPath, exist(pluginPath));
    end


    required_plugin_files = [ 
	"communicatingJobWrapper.sh",
	"communicatingJobWrapperSmpd.sh",
	"independentJobWrapper.sh",
    ];
    for k = numel(required_plugin_files)
	tmp1 = required_plugin_files(k);
	tmp2 = fullfile(pluginPath, tmp1);
	if exist(tmp2) ~= 2
	    error( errorIDroot + "BadPluginPath", ...
	        "Plugin path %s missing required file %s", ...
		pluginPath, tmp1);
	end
    end


    %-------- Get the RemoteJobStorageLocation
    % Generate default value
    tmp1 = sprintf( cluster_jobstorage_string, username);
    remJobStorageLoc = configClusterArgumentQuery( ...
        passed = Args.remoteJobStorageLocation, ...
	defaultValue = tmp1, ...
	interactive = interactive, ...
	prompt = "Where should MATLAB store its job files on the cluster?", ...
	type = "string" ...
    );
    %if exist(tmp1) ~= 2
    %    error( errorIDroot + "BadRemJobStorageLoc", ...
    %           "Invalid value %s for remoteJobStorageLocation, not " + ...
    %           "a directory; %d", remJobStorageLoc, exist(tmp1));
    %end

    %--------  Name of profile to save as
    % default value
    tmp1 = sprintf("%s HPC", ClusterName);
    profileName = configClusterArgumentQuery( ...
        passed = Args.profileName, ...
        defaultValue = tmp1, ...
	interactive = interactive, ...
	prompt = "What should this profile be named?", ...
	type = "string" ...
    );

    % Make sure no existing profile with this name exits
    existing_profiles = parallel.clusterProfiles;
    for k = 1:numel(existing_profiles)
	existing_prof = existing_profiles{k};
	if strcmp(profileName, existing_prof) == 1
	    error( errorIDroot + "DuplicateProfile", ...
	        "A profile named '%s' already exists, and we refuse to " + ...
		"overwrite it", profileName);
	end
    end

    % Do not know how to set this
    if useLocalCluster
        profileDesc = sprintf("Profile for use locally on HPC cluster %s", ...
	    ClusterName);
    else
	profileDesc = sprintf("Profile for use on your local " + ...
	    "workstation, sending jobs to the %s HPC cluster", ClusterName);
    end

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %		Build cluster profile
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    profile = parallel.cluster.Generic;
    %profile.Description = profileDesc;

    %profile.Profile = profile_name; %Cant set, read-only
    profile.ClusterMatlabRoot = cluster_matlab_path;
    profile.OperatingSystem = "unix";
    profile.HasSharedFilesystem = useLocalCluster;
    profile.RequiresOnlineLicensing = boolean(0);
    profile.PluginScriptsLocation = pluginPath;
    %profile.AutoAttachFiles = boolean(1)

    % These should not be needed when useLocalCluster, but should
    % not hurt either
    profile.AdditionalProperties.AuthenticationMode = "Multifactor";
    profile.AdditionalProperties.ClusterHost = cluster_login_node;
    profile.AdditionalProperties.UseIdentityFile = 0;
    profile.AdditionalProperties.Username = username;

    profile.AdditionalProperties.RemoteJobStorageLocation = remJobStorageLoc;
    profile.AdditionalProperties.UseUniqueSubfolders = boolean(1);


    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %		Confirmation
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    fprintf("About to create profile:\n");
    fprintf("Profile name: %s\n", profileName);
    %printf("Description: %s\n", profileDesc);
    profile
    profile.AdditionalProperties

    if interactive && Args.requireConfirmation
        confirm = configClusterArgumentQuery( ...
	    interactive = interactive, ...
	    prompt = "Shall we proceed to create the profile? [y/N] ", ...
            type = "string" ...
	);
        if isempty(confirm)
	    confirm = logical(0);
	elseif startsWith(lower(confirm), "n")
	    confirm = logical(0);
	elseif startsWith(lower(confirm), "y")
	    confirm = logical(1);
	else
	    error( errorIDroot + "InvalidValue", ...
	        "Invalid value '%s' for confirm, expecting n or y");
	end

	if ~confirm 
            error( errorIDroot + "NotConfirmed", ...
	        "Aborting due to user not confirming");
        end
    end
    profile.saveAsProfile(profileName);
    cluster = profile
end








